import traitlets
from nbconvert import preprocessors


class LiteraryPythonPreprocessor(preprocessors.Preprocessor):
    default_include_tags = traitlets.Set(
        traitlets.Unicode(), default_value={"export", "docstring"}
    )

    def preprocess(self, nb, resources):
        nb.cells = [*self.iter_permitted_cells(nb, resources)]
        return nb, resources

    def iter_permitted_cells(self, nb, resources):
        for i, cell in enumerate(nb.cells):
            tags = set(cell.metadata.get("tags", []))

            if self.default_include_tags & tags:
                yield cell



import functools
import pathlib
from typing import Any, Dict

import toml


@functools.lru_cache()
def find_project_path(*search_paths) -> pathlib.Path:
    """Find the root path of the literary project.

    Searches for known files / directories including
    `.git/`, `.hg/`, or `pyproject.toml`.

    :param search_paths: starting search paths
    :return:
    """
    if not search_paths:
        search_paths = [pathlib.Path.cwd()]

    visited = set()

    for search_path in search_paths:
        for path in (search_path, *search_path.parents):

            # Avoid re-visiting paths
            if path in visited:
                break
            visited.add(path)

            if (path / ".git").exists():
                return path

            if (path / ".hg").exists():
                return path

            if (path / "pyproject.toml").exists():
                return path

    raise FileNotFoundError("Couldn't find project path")


def load_pyproject_config(path: pathlib.Path) -> Dict[str, Any]:
    """Load configuration from a pyproject.toml file, modifying the various parameters
    where appropriate.

    :param path: path to pyproject.toml file
    :return:
    """
    data = toml.load(path)
    project_path = path.parent.absolute()

    config = {
        k.replace("-", "_"): v
        for k, v in data.get("tool", {}).get("literary", {}).items()
    }

    if (source_path := config.get("source_path")) is not None:
        source_path = (project_path / source_path).resolve()

    if (package_path := config.get("package_path")) is not None:
        package_path = (project_path / package_path).resolve()

    if test_paths := config.get("test_paths", []):
        test_paths = [(project_path / p).resolve() for p in test_paths]

    test_processes = config.get("test_processes")

    return {
        "source_path": source_path,
        "package_path": package_path,
        "test_paths": test_paths,
        "test_processes": test_processes,
    }


def load_default_config(project_path: pathlib.Path) -> Dict[str, Any]:
    """Load configuration from a pyproject.toml file, returning an empty
    dictionary if the file does not exist

    :param project_path: path to pyproject.toml file
    :return:
    """
    pyproject_path = project_path / "pyproject.toml"
    if pyproject_path.exists():
        return load_pyproject_config(pyproject_path)
    return {}



from typing import Literal


def escape_triple_quotes(string: str, single_quote: Literal["'", '"'] = '"') -> str:
    """Escape triple quotes inside a string

    :param string: string to escape
    :param single_quote: single-quote character
    :return: escaped string
    """
    assert len(single_quote) == 1
    quote = "".join([single_quote] * 3)

    escaped_single_quote = rf"\{single_quote}"
    escaped_quote = "".join([escaped_single_quote] * 3)
    return string.replace(quote, escaped_quote)



import ast
import sys
import logging

import traitlets.config
from jinja2 import DictLoader
from nbconvert import exporters
from traitlets import default, import_item, List

from .preprocessor import LiteraryPythonPreprocessor
from .utils import escape_triple_quotes

logger = logging.getLogger(__name__)

# Backwards compatibility
if sys.version_info < (3, 9, 0):
    import astunparse
    import astunparse.unparser

    class ASTUnparser(astunparse.unparser.Unparser):
        """AST unparser with additional preference for triple-quoted multi-line strings"""

        def _Constant(self, tree):
            if isinstance(tree.value, str) and "\n" in tree.value:
                self.write(f'"""{escape_triple_quotes(tree.value)}"""')
                return

            super()._Constant(tree)

    # Monkey patch to ensure correctness
    astunparse.Unparser = ASTUnparser
    astunparse.unparser.Unparser = ASTUnparser

    from astunparse import unparse as unparse_ast

else:
    from ast import unparse as unparse_ast


class LiteraryPythonExporter(exporters.PythonExporter):
    def __init__(self, *args, **kwargs):
        """
        Public constructor

        Parameters
        ----------
        config : ``traitlets.config.Config``
            User configuration instance.
        `**kw`
            Additional keyword arguments passed to parent __init__

        """

        super().__init__(*args, **kwargs)

        self._init_transformers()

    transformers = List(
        [
            "literary.transformers.PatchTransformer",
            "literary.transformers.IPythonTransformer",
        ]
    ).tag(config=True)

    def _init_transformers(self):
        self._transformers = []

        for value in self.transformers:
            if isinstance(value, str):
                value = import_item(value)
            self._transformers.append(value())

    @default("template_file")
    def _template_file_default(self):
        return "literary"

    @default("extra_loaders")
    def _extra_loaders_default(self):
        loader = DictLoader(
            {
                "literary": '''
{%- extends 'python/index.py.j2' -%}

{% block header %}
"""
{%- for cell in nb.cells -%}
{%- if "docstring" in cell.metadata.tags -%}
{{ cell.source | escape_triple_quotes }}
{% endif -%}
{%- endfor -%}
"""
{% endblock header %}

{%- block any_cell scoped -%}
{%- if "export" in cell.metadata.tags -%}
{{ super() }}
{%- endif -%}
{%- endblock any_cell -%}

                             '''
            }
        )
        return [loader]

    @default("default_preprocessors")
    def _default_preprocessors_default(self):
        return [LiteraryPythonPreprocessor]

    @default("exclude_input_prompt")
    def _exclude_input_prompt_default(self):
        return True

    @property
    def default_config(self):
        c = traitlets.config.Config(
            {
                "LiteraryPythonPreprocessor": {"enabled": True},
            }
        )
        c.merge(super().default_config)
        return c

    def default_filters(self):
        yield from super().default_filters()
        yield "escape_triple_quotes", escape_triple_quotes

    def from_notebook_node(self, nb, resources=None, **kwargs):
        body, resources = super().from_notebook_node(nb, resources, **kwargs)
        node = ast.parse(body)

        try:
            # Support ast.NodeTransformer and custom transformers using
            # `visit()` API
            for transformer in self._transformers:
                node = transformer.visit(node)
        except Exception as err:
            raise RuntimeError(
                f"An error occurred during AST transforming: {body}"
            ) from err

        return unparse_ast(node), resources
